import getpass
import http
import logging
import os
import posixpath
from typing import List, Optional, Union
import urllib.parse

import requests
# For backward compatible HTTPError
import urllib3.exceptions

from .config import Config, parseClassName
from .config import ConfigOptions
from .config import parseURLDomain
from .constants import *
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from .identified import Identified
from . import SearchQuery


class PartShop:
    """A class which provides an API front-end for
    online bioparts repositories"""

    def __init__(self, url, spoofed_url=''):
        """

        :param url: The URL of the online repository (as a str)
        :param spoofed_url:
        """
        # initialize member variables
        self.resource = self._validate_url(url, 'resource')
        self.user = ''
        self.key = ''
        self.spoofed_resource = self._validate_url(spoofed_url, 'spoofed')

    def _validate_url(self, url, url_name):
        """ Function to validate url with type checking, urlparse and
        terminal forward slash's as well as allowing spoofed url to
        pass as ''
        """
        # If url_name = spoofed and url = '' pass, as spoofed_resource
        # initialises with default spoofed_url = '' when objects are
        # created, ideal behavior should not throw errors in such case.
        if url_name == "spoofed" and url == '':
            return url
        # Type check to ensure passed url is a string
        if not isinstance(url, str):
            msg = ('PartShop initialization failed. The {} URL '
                   + 'is not of type string').format(url_name)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        # Authenticity check with urlparse as to ensure correct scheme and
        # netloc present
        url_pieces = urllib.parse.urlparse(url)
        if not all([url_pieces.scheme in ['http', 'https'],
                   url_pieces.netloc]):
            msg = ('PartShop initialization failed. The {} URL '
                   + 'was not valid').format(url_name)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        # String check to ensure length > 0 and does not contain terminal "/"
        if len(url) > 0 and url[-1] == '/':
            msg = ('PartShop initialization failed. The {} URL '
                   + 'should not contain a terminal forward slash')
            msg = msg.format(url_name)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        return url

    @property
    def logger(self):
        logger = logging.getLogger('sbol2')
        if not logger.hasHandlers():
            # If there are no handlers, nobody has initialized
            # logging.  Configure logging here so we have a chance of
            # seeing the messages.
            logging.basicConfig()
        return logger

    def count(self):
        """Return the count of objects contained in a PartShop"""
        raise NotImplementedError('Not yet implemented')

    def spoof(self, spoofed_url):
        self.spoofed_resource = self._validate_url(spoofed_url, 'spoofed')

    def sparqlQuery(self, query):
        """
        Issue a SPARQL query
        :param query: the SPARQL query
        :return: the HTTP response object
        """
        endpoint = parseURLDomain(self.resource) + '/sparql'
        if self.spoofed_resource == '':
            resource = self.resource
        else:
            resource = self.spoofed_resource
        p = query.find('WHERE')
        if p != -1:
            from_clause = ' FROM <' + parseURLDomain(resource) + \
                          '/user/' + self.user + '> '
            query = query[:p].rstrip() + from_clause + query[p:].lstrip()
        headers = {'X-authorization': self.key, 'Accept': 'application/json'}
        params = {'query': query}  # should handle encoding the query
        if Config.getOption(ConfigOptions.VERBOSE.value) is True:
            self.logger.debug('Issuing SPARQL: ' + query)
        response = requests.get(endpoint, headers=headers, params=params)
        if not response:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST,
                            response)
        return response

    def pull(self, uris, doc, recursive=True):
        """Retrieve an object from an online resource
        :param uris: A list of SBOL objects you want to retrieve,
        or a single SBOL object URI
        :param doc: A document to add the data to
        :param recursive: Whether the GET request should be recursive
        :return: nothing (doc parameter is updated, or an exception is thrown)
        """
        # IMPLEMENTATION NOTE: rdflib.Graph.parse() actually lets you
        # pass a URL as an argument. I decided to not use this method,
        # because I couldn't find an easy way to get the response
        # code, set HTTP headers, etc. In addition, I would need
        # to use requests for submitting new SBOL data anyway.
        endpoints = []
        if type(uris) is str:
            endpoints.append(uris)
        elif type(uris) is list:
            endpoints = uris
        else:
            raise TypeError('URIs must be str or list. Found: ' + str(type(uris)))
        for uri in endpoints:
            try:
                query = self._uri2url(uri)
            except SBOLError as err:
                if err.error_code() == SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT:
                    # Assume user has only specified displayId
                    query = self.resource + '/' + uri
                else:
                    raise
            query += '/sbol'
            if not recursive:
                query += 'nr'
            if Config.getOption(ConfigOptions.VERBOSE.value):
                self.logger.debug('Issuing GET request ' + query)
            # Issue GET request
            response = requests.get(query,
                                    headers={'X-authorization': self.key,
                                             'Accept': 'text/plain'})
            if response.status_code == 404:
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                                'Part not found. Unable to pull: ' + query)
            elif response.status_code == 401:
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_HTTP_UNAUTHORIZED,
                                'Please log in with valid credentials')
            elif not response:
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, response)
            # Add content to document
            doc.appendString(response.content, overwrite=True)
            doc.resource_namespaces.add(self.resource)

    def submit(self, doc, collection='', overwrite=0):
        """Submit a SBOL Document to SynBioHub
        :param doc: The Document to submit
        :param collection: The URI of a SBOL Collection to which the Document
        contents will be uploaded
        :param overwrite: An integer code: 0 (default) - do not overwrite,
        1 - overwrite, 2 - merge
        :return: the HTTP response object
        """
        if collection == '':
            # If a Document is submitted as a new collection,
            # then Document metadata must be specified
            if len(doc.displayId) == 0 or len(doc.name) == 0 \
                    or len(doc.description) == 0:
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT,
                                'Cannot submit Document. The Document must be '
                                'assigned a displayId, name, and ' +
                                'description for upload.')
        else:
            if len(self.spoofed_resource) > 0 and self.resource in collection:
                # Correct collection URI in case a spoofed resource is being used
                collection = collection.replace(self.resource,
                                                self.spoofed_resource)
            if Config.getOption(ConfigOptions.VERBOSE.value) is True:
                self.logger.info('Submitting Document to an existing collection: %s',
                                 collection)
        # if Config.getOption(ConfigOptions.SERIALIZATION_FORMAT.value) == 'rdfxml':
        #     self.addSynBioHubAnnotations(doc)
        files = {}
        if len(doc.displayId) > 0:
            files['id'] = (None, doc.displayId)
        if len(doc.version) > 0:
            files['version'] = (None, doc.version)
        if doc.name and len(doc.name) > 0:
            files['name'] = (None, doc.name)
        if doc.description and len(doc.description) > 0:
            files['description'] = (None, doc.description)
        citations = ''
        for citation in doc.citations:
            citations += citation + ','
        citations = citations[0:-1]  # chop off final comma
        files['citations'] = (None, citations)
        keywords = ''
        for kw in doc.keywords:
            keywords += kw + ','
        keywords = keywords[0:-1]
        files['keywords'] = (None, keywords)
        files['overwrite_merge'] = (None, str(overwrite))
        files['user'] = (None, self.key)
        files['file'] = ('file', doc.writeString(), 'text/xml')
        if collection != '':
            files['rootCollections'] = (None, collection)
        # Send POST request
        # print(files)
        response = requests.post(self.resource + '/submit',
                                 files=files,
                                 headers={'Accept': 'text/plain',
                                          'X-authorization': self.key})
        # print(response.text)
        if response:
            return response
        elif response.status_code == 401:
            # Raise a urllib3 HTTPError exception to be backward compatible with pySBOL
            raise urllib3.exceptions.HTTPError('You must login with valid credentials '
                                               'before submitting')
        else:
            # Raise a urllib3 HTTPError exception to be backward compatible with pySBOL
            raise urllib3.exceptions.HTTPError('HTTP post request failed with: ' +
                                               str(response.status_code) +
                                               ' - ' + str(response.content))

    def _uri2url(self, uri):
        """Converts an SBOL URI to a URL for running queries to a SynBioHub
        endpoint.

        """
        if self.resource in uri:
            return uri
        if parseURLDomain(self.resource) in uri:
            return uri
        if self.spoofed_resource and self.spoofed_resource in uri:
            return uri.replace(self.spoofed_resource, self.resource)
        msg = ('{} does not exist in the resource namespace')
        msg = msg.format(uri)
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)

    def remove(self, uri):
        query = self._uri2url(uri)
        url = '{}/remove'.format(query)
        headers = {
            'X-authorization': self.key,
            'Accept': 'application/json'
        }
        response = requests.get(url, headers=headers)
        if response.ok:
            return True
        if response.status_code == requests.codes.unauthorized:
            # Handle a 401 Unauthorized error
            msg = 'You must login with valid credentials before removing'
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_HTTP_UNAUTHORIZED, msg)
        # Not sure what went wrong
        msg = 'Unknown error: ' + response
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)

    def login(self, user_id, password=''):
        """In order to submit to a PartShop, you must login first.
        Register on [SynBioHub](http://synbiohub.org) to
        obtain account credentials.
        :param user_id: User ID
        :param password: User password
        :return: the HTTP response object
        """
        self.user = user_id
        if password is None or password == '':
            password = getpass.getpass()
        response = requests.post(
            parseURLDomain(self.resource) + '/remoteLogin',
            data={'email': user_id, 'password': password},
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )
        if not response:
            msg = 'Login failed due to an HTTP error: {}'
            msg = msg.format(response)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)
        self.key = response.content.decode('utf-8')
        return response

    def getKey(self):
        return self.key

    def getURL(self):
        return self.resource

    # For backward compatibility with pySBOL
    def getUser(self):
        return self.user

    # For backward compatibility with pySBOL
    def getSpoofedURL(self):
        return self.spoofed_resource

    # def addSynBioHubAnnotations(self, doc):
    #     doc.addNamespace("http://wiki.synbiohub.org/wiki/Terms/synbiohub#", "sbh")
    #     for id, toplevel_obj in doc.SBOLObjects:
    #         toplevel_obj.apply(None, None)  # TODO

    def attachFile(self, top_level_uri, filepath):
        """Attach a file to an object in SynBioHub.

        Returns None if successful.

        Raises SBOLError with code SBOL_ERROR_HTTP_UNAUTHORIZED if it
        there is an HTTP Unauthorized response.

        Raises SBOLError with code SBOL_ERROR_BAD_HTTP_REQUEST on any
        other HTTP error. The actual status code is embedded in the
        string message.

        """
        # Expand User
        filepath = os.path.expanduser(filepath)
        # HTTP request headers
        headers = {
            'Accept': 'text/plain',
            'X-authorization': self.key
        }
        url = posixpath.join(top_level_uri, 'attach')
        with open(filepath, 'rb') as fp:
            files = {'file': fp}
            response = requests.post(url, headers=headers, files=files)

        if response.ok:
            if Config.getOption(ConfigOptions.VERBOSE.value) is True:
                print(response.text)
            return
        if response.status_code == http.HTTPStatus.UNAUTHORIZED:
            # HTTP 401
            msg = 'You must login with valid credentials before attaching a file'
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_HTTP_UNAUTHORIZED, msg)
        # Not sure what went wrong
        msg = 'HTTP Error code {} trying to attach file.'
        msg = msg.format(response.status_code)
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)

    def downloadAttachment(self, attachment_uri, filepath='.'):
        '''Download a file attachment from SynBioHub.
        :param attachment_uri: The URI of the file to download
        :param filepath: The local path or filename to which the file will be written
        Returns None if successful.

        Raises EnvironmentError if the filepath is invalid

        Raises SBOLError with code SBOL_ERROR_HTTP_UNAUTHORIZED if it
        there is an HTTP Unauthorized response.

        Raises SBOLError with code SBOL_ERROR_BAD_HTTP_REQUEST on any
        other HTTP error. The actual status code is embedded in the
        string message.
        '''
        url = self._uri2url(attachment_uri)
        url = posixpath.join(url, 'download')
        filepath = os.path.expanduser(filepath)

        # HTTP request headers
        headers = {
            'Accept': 'text/plain',
            'X-authorization': self.key
        }

        # Issue GET request
        response = requests.get(url, headers=headers)
        if response.ok:
            filename = response.headers['Content-Disposition']
            filename = filename[22:-1]  # remove extraneous text: attachment; filename="
            if os.path.isdir(filepath):
                filepath = posixpath.join(filepath, filename)
            with open(filepath, 'wb') as filehandle:
                filehandle.write(response.content)
            return
        if response.status_code == http.HTTPStatus.UNAUTHORIZED:
            # HTTP 401
            msg = 'You must login with valid credentials before downloading a file'
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_HTTP_UNAUTHORIZED, msg)
        if response.status_code == http.HTTPStatus.NOT_FOUND:
            # HTTP 404
            msg = 'Unable to download. Attachment {} not found.'.format(attachment_uri)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND, msg)

        # Not sure what went wrong
        msg = 'HTTP Error code {} trying to download file.'
        msg = msg.format(response.status_code)
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)

    def _make_search_item(self, item: dict) -> Identified:
        obj = Identified()
        obj.identity = item['uri']
        obj.displayId = item['displayId']
        obj.name = item['name']
        obj.description = item['description']
        obj.version = item['version']
        return obj

    def _search(self, url: str) -> List[Identified]:
        """Given a URL, perform the search and parse the results.
        The URL is formed by one of the other search methods: search_exact,
        search_general, and search_advanced.
        """
        # Login is optional
        headers = {'Accept': 'text/plain'}
        if self.key:
            headers['X-authorization'] = self.key

        self.logger.info('search query: %s', url)
        response = requests.get(url, headers=headers)
        if not response:
            # Something went wrong
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, response)
        # Everything looks good, parse and return the results
        return [self._make_search_item(item) for item in (response.json())]

    def search_general(self, search_text: str,
                       object_type: Optional[str] = SBOL_COMPONENT_DEFINITION,
                       offset: int = 0, limit: int = 25) -> List[Identified]:
        # See https://synbiohub.github.io/api-docs/#search-metadata
        search_url = parseURLDomain(self.resource)
        query = dict(objectType=parseClassName(object_type))
        query = urllib.parse.urlencode(query)
        search_text = urllib.parse.quote(search_text)
        params = dict(offset=offset, limit=limit)
        params = urllib.parse.urlencode(params)
        query_url = f'{search_url}/search/{query}&{search_text}/?{params}'
        return self._search(query_url)

    def search_exact(self, search_text: str,
                     object_type: Optional[str] = SBOL_COMPONENT_DEFINITION,
                     property_uri: Optional[str] = None,
                     offset: int = 0, limit: int = 25) -> List[Identified]:
        # See https://synbiohub.github.io/api-docs/#search-metadata
        search_url = parseURLDomain(self.resource)
        query = dict(objectType=parseClassName(object_type))
        if search_text.startswith('http'):
            search_text = f"<{search_text}>"
        else:
            search_text = f"'{search_text}'"
        query[parseClassName(property_uri)] = search_text
        query = urllib.parse.urlencode(query)
        params = dict(offset=offset, limit=limit)
        params = urllib.parse.urlencode(params)
        query_url = f'{search_url}/search/{query}&/?{params}'
        return self._search(query_url)

    def search_advanced(self, search_query: SearchQuery):
        # See https://synbiohub.github.io/api-docs/#search-metadata
        search_url = parseURLDomain(self.resource)
        query = search_query.query_dict()
        # if search_text.startswith('http'):
        #     search_text = f"<{search_text}>"
        # else:
        #     search_text = f"'{search_text}'"
        # query[parseClassName(property_uri)] = search_text
        query = urllib.parse.urlencode(query)
        params = dict(offset=(search_query.offset),
                      limit=(search_query.limit))
        params = urllib.parse.urlencode(params)
        query_url = f'{search_url}/search/{query}&/?{params}'
        return self._search(query_url)

    def search(self, search_text: Union[str, SearchQuery],
               object_type: Optional[str] = SBOL_COMPONENT_DEFINITION,
               property_uri: Optional[Union[str, int]] = None,
               offset: int = 0, limit: int = 25) -> List[Identified]:
        # if search_text is a SearchQuery, dispatch to search_advanced
        if type(search_text) is SearchQuery:
            return self.search_advanced(search_text)
        if type(property_uri) is int:
            # User called without property_uri and with offset. Shift args
            if offset > 0:
                limit = offset
            offset = property_uri
            property_uri = None
        # if property_uri is not specified, do a general search
        if property_uri is None:
            return self.search_general(search_text, object_type, offset,
                                       limit)
        # property_uri is specified, do an exact search
        return self.search_exact(search_text, object_type, property_uri,
                                 offset, limit)

    def _search_count(self, url):
        headers = {'Accept': 'text/plain'}
        if self.key:
            headers['X-authorization'] = self.key
        self.logger.info('searchCount query: %s', url)
        response = requests.get(url, headers=headers)
        if not response:
            # Something went wrong
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, response)
        # Everything looks good, parse and return the results
        return int(response.text)

    def search_count(self, search_text, object_type, property_uri):
        search_url = parseURLDomain(self.resource)
        query = dict(objectType=parseClassName(object_type))
        query = urllib.parse.urlencode(query)
        search_text = urllib.parse.quote(search_text)
        # params = dict(offset=offset, limit=limit)
        # params = urllib.parse.urlencode(params)
        # query_url = f'{search_url}/search/{query}&{search_text}/?{params}'
        url = f'{search_url}/searchCount/{query}&{search_text}/'
        return self._search_count(url)

    def search_count_advanced(self, search_query):
        search_url = parseURLDomain(self.resource)
        query = search_query.query_dict()
        query = urllib.parse.urlencode(query)
        url = f'{search_url}/searchCount/{query}&/'
        return self._search_count(url)

    def searchCount(self, search_text, object_type=None, property_uri=None):
        """Returns the number of records matching the given criteria.
        """
        # if search_text is a SearchQuery, dispatch to search_count_advanced
        if type(search_text) is SearchQuery:
            return self.search_count_advanced(search_text)

        # if object_type is not specified, default to SBOL_COMPONENT_DEFINITION
        if object_type is None:
            object_type = SBOL_COMPONENT_DEFINITION

        # Dispatch to the internal search count method
        return self.search_count(search_text, object_type, property_uri)



import os
import unittest

import sbol2 as sbol

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
TEST_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'crispr_example.xml')


class TestConstants(unittest.TestCase):

    def test_igem_standard_assembly(self):
        # IGEM_STANDARD_ASSEMBLY is a function in pySBOL.
        self.assertTrue(callable(sbol.IGEM_STANDARD_ASSEMBLY),
                        'IGEM_STANDARD_ASSEMBLY is not callable')

    def test_in_roles(self):
        doc = sbol.Document()
        doc.read(TEST_LOCATION)
        md_uri = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0'
        md = doc.moduleDefinitions[md_uri]
        i_uri = ('http://sbols.org/CRISPR_Example/CRISPR_Template' +
                 '/cas9_complex_formation/1.0.0')
        i = md.interactions[i_uri]
        p_uri = ('http://sbols.org/CRISPR_Example/CRISPR_Template' +
                 '/cas9_complex_formation/cas9_gRNA_complex/1.0.0')
        p = i.participations[p_uri]
        # p.roles is a list of URIs
        # The constant should be in the list of roles
        self.assertIn(sbol.SBO_PRODUCT, p.roles)



import time
import timeit
import sbol2

EXAMPLE_PATH = '../test/SBOLTestSuite/SBOL2/RepressionModel.xml'
EXAMPLE_PATH = '../W303.xml'


def load_document(doc, path, n_times=1):
    for i in range(n_times):
        doc = sbol2.Document(path)
    return doc


start = time.time()
my_doc = load_document(None, EXAMPLE_PATH, 1)
end = time.time()
print('elapsed seconds: {}'.format(end-start))
print('Loaded document size: {}'.format(len(my_doc)))



import urllib.parse

import requests

search_text = 'NAND'
offset = 0
limit = 10

# search_url = parseURLDomain(self.resource)
search_url = 'https://synbiohub.org'

# query = dict(objectType=parseClassName(object_type))
query = dict(objectType='ComponentDefinition')
query = urllib.parse.urlencode(query)
search_text = urllib.parse.quote(search_text)
params = dict(offset=offset, limit=limit)
params = urllib.parse.urlencode(params)
query_url = f'{search_url}/search/{query}&{search_text}/?{params}'
print(query_url)

# return self._search(query_url)
headers = {'Accept': 'text/plain'}
# if self.key:
#     headers['X-authorization'] = self.key

# self.logger.info('search query: %s', url)
response = requests.get(query_url, headers=headers)
if not response:
    # Something went wrong
    print('Failure')
print(response.json())



import sbol2

sbol2.setHomespace('http://me.name/test#')
seq = sbol2.Sequence('seq', 'agct')

cmp = sbol2.ComponentDefinition('cmp')
cmp.sequences = seq.identity

prom = sbol2.ComponentDefinition('prom')
prom.roles = [sbol2.SO_PROMOTER]
prom.displayId = "p1"
prom.name = "Promoter 1"
prom.description = "The first promoter"

doc = sbol2.Document()
doc.add_list([seq, cmp, prom])
doc.write('shortbol.xml')



# Upload an SBOL file to SynBioHub

import argparse

import sbol2


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("sbol", metavar="SBOL_FILE")
    parser.add_argument("url", metavar="PARTSHOP_URL")
    parser.add_argument('-d', '--description', default='test description')
    parser.add_argument('-n', '--name', default='test data')
    parser.add_argument('-i', '--id', default='test')
    parser.add_argument('-u', '--user', default='testuser')
    parser.add_argument('-p', '--password', default='test')
    args = parser.parse_args(args)
    return args


def main(argv=None):
    args = parse_args(argv)

    doc = sbol2.Document(args.sbol)
    doc.description = args.description
    doc.displayId = args.id
    doc.name = args.name

    part_shop = sbol2.PartShop(args.url)
    # part_shop.spoof('https://synbiohub.org')
    part_shop.login(args.user, args.password)
    part_shop.submit(doc)


if __name__ == '__main__':
    main()



#   Copyright (C) 2018 pysbolgraph developers. All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions
#   are met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
#   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
#   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
from collections import defaultdict
from typing import Dict

from lxml import etree
from lxml.etree import tostring
from lxml.etree import QName

from rdflib.namespace import RDF
from rdflib import URIRef, Literal

rdfNS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
sbolNS = "http://sbols.org/v2#"

# This is a module global that is used to register ownership relationships
# between classes. The serializer uses this to generate structured XML from
# flat RDF/XML
OWNERSHIP_PREDICATES: Dict[URIRef, set] = defaultdict(set)


def is_ownership_relation(triple, subject_type):
    # subject = triple[0]
    predicate = triple[1]
    # object = triple[2]

    return (predicate in OWNERSHIP_PREDICATES and
            subject_type in OWNERSHIP_PREDICATES[predicate])


def register_ownership_relation(parent_type, predicate):
    # Encapsulates and populates the OWNERSHIP_PREDICATES dictionary.
    #
    # :param parent_type: The RDF type of the parent class
    # :param predicate: The URI for the property.
    OWNERSHIP_PREDICATES[URIRef(predicate)].add(URIRef(parent_type))


def ns_prefix_dict(g):
    """Return a dictionary of namespace, uri prefix pairs."""
    return {ns: prefix.toPython() for (ns, prefix) in g.namespaces()}


def serialize_sboll2(g):
    prefixes = ns_prefix_dict(g)
    prefixes['rdf'] = rdfNS
    prefixes['sbol'] = sbolNS

    subject_to_element = dict()
    subject_to_type = dict()

    owned_elements = set()

    for triple in g.triples((None, RDF.type, None)):
        subject_to_type[triple[0]] = triple[2]
        subject = triple[0].toPython()
        the_type = triple[2].toPython()
        if subject in subject_to_element:
            etree.SubElement(subject_to_element[subject],
                             prefixify(RDF.type, prefixes, True),
                             attrib={
                                 QName(rdfNS, 'resource'): the_type
                             })
        else:
            subject_to_element[subject] = etree.Element(prefixify(the_type, prefixes,
                                                                  True),
                                                        attrib={
                                                            QName(rdfNS,
                                                                  'about'): subject
                                                        }
                                                        )

    for triple in g.triples((None, None, None)):
        if triple[1] == RDF.type:
            continue
        subject = triple[0].toPython()
        predicate = triple[1].toPython()
        obj = triple[2]
        element = subject_to_element[subject]
        if is_ownership_relation(triple, subject_to_type[triple[0]]):
            owned_element = subject_to_element[obj.toPython()]
            ownership_element = etree.SubElement(element,
                                                 prefixify(predicate, prefixes, True))
            ownership_element.append(owned_element)
            owned_elements.add(obj.toPython())
            continue
        if isinstance(obj, URIRef):
            etree.SubElement(element, prefixify(predicate, prefixes, True), attrib={
                QName(rdfNS, 'resource'): obj.toPython()
            })
        elif isinstance(obj, Literal):
            elem = etree.SubElement(element, prefixify(predicate, prefixes, True))
            elem.text = obj
        else:
            raise Exception()

    doc = etree.Element(QName(rdfNS, 'RDF'), nsmap=prefixes)

    for subject in subject_to_element:
        if subject not in owned_elements:
            element = subject_to_element[subject]
            doc.append(element)

    return tostring(doc, pretty_print=True)


def prefixify(iri, prefixes, create_new):
    for prefix in prefixes:
        prefix_iri = prefixes[prefix]
        if iri.startswith(prefix_iri):
            return QName(prefix_iri, iri[len(prefix_iri):])
    if not create_new:
        return iri
    fragment_start = iri.rfind('#')
    if fragment_start == -1:
        fragment_start = iri.rfind('/')
    if fragment_start == -1:
        return iri
    iri_prefix = iri[:fragment_start + 1]
    i = 0
    while True:
        prefix_name = 'ns' + str(i)
        if prefix_name not in prefixes:
            prefixes[prefix_name] = iri_prefix
            return QName(iri_prefix, iri[len(iri_prefix):])
        i = i + 1



import unittest

import sbol2


class TestError(unittest.TestCase):

    def test_sbol_error_warning(self):
        # SBOLError should raise a TypeError if the first argument
        # is not an SBOLErrorCode.
        with self.assertRaises(TypeError):
            sbol2.SBOLError('Item not found',
                            sbol2.SBOLErrorCode.NOT_FOUND_ERROR)

    def test_what(self):
        msg = 'Item not found'
        error = sbol2.SBOLError(sbol2.SBOLErrorCode.NOT_FOUND_ERROR, msg)
        self.assertEqual(error.what(), msg)

    def test_error_code(self):
        msg = 'Item not found'
        code = sbol2.SBOLErrorCode.NOT_FOUND_ERROR
        error = sbol2.SBOLError(code, msg)
        self.assertEqual(error.error_code(), code)



from enum import Enum
from typing import Any
import warnings

from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from rdflib import URIRef


class FileFormats(Enum):
    JSON = 'json'
    NTRIPLES = 'ntriples'
    RDFXML = 'rdfxml'


class ConfigOptions(Enum):
    HOMESPACE = 'homespace'
    SBOL_COMPLIANT_URIS = 'sbol_compliant_uris'
    SBOL_TYPED_URIS = 'sbol_typed_uris'
    SERIALIZATION_FORMAT = 'serialization_format'
    VALIDATE = 'validate'
    VALIDATOR_URL = 'validator_url'
    LANGUAGE = 'language'
    TEST_EQUALITY = 'test_equality'
    CHECK_URI_COMPLIANCE = 'check_uri_compliance'
    CHECK_COMPLETENESS = 'check_completeness'
    CHECK_BEST_PRACTICES = 'check_best_practices'
    FAIL_ON_FIRST_ERROR = 'fail_on_first_error'
    PROVIDE_DETAILED_STACK_TRACE = 'provide_detailed_stack_trace'
    URI_PREFIX = 'uri_prefix'
    SUBSET_URI = 'subset_uri'
    VERSION = 'version'
    INSERT_TYPE = 'insert_type'
    MAIN_FILE_NAME = 'main_file_name'
    DIFF_FILE_NAME = 'diff_file_name'
    RETURN_FILE = 'return_file'
    VERBOSE = 'verbose'


options = {
    ConfigOptions.HOMESPACE.value: 'http://examples.org',
    ConfigOptions.SBOL_COMPLIANT_URIS.value: True,
    ConfigOptions.SBOL_TYPED_URIS.value: True,
    ConfigOptions.SERIALIZATION_FORMAT.value: 'sbol',
    ConfigOptions.VALIDATE.value: True,
    ConfigOptions.VALIDATOR_URL.value: 'https://validator.sbolstandard.org/validate/',
    ConfigOptions.LANGUAGE.value: 'SBOL2',
    ConfigOptions.TEST_EQUALITY.value: False,
    ConfigOptions.CHECK_URI_COMPLIANCE.value: False,
    ConfigOptions.CHECK_COMPLETENESS.value: False,
    ConfigOptions.CHECK_BEST_PRACTICES.value: False,
    ConfigOptions.FAIL_ON_FIRST_ERROR.value: False,
    ConfigOptions.PROVIDE_DETAILED_STACK_TRACE.value: False,
    ConfigOptions.URI_PREFIX.value: '',
    ConfigOptions.SUBSET_URI.value: '',
    ConfigOptions.VERSION.value: '',
    ConfigOptions.INSERT_TYPE.value: False,
    ConfigOptions.MAIN_FILE_NAME.value: 'main file',
    ConfigOptions.DIFF_FILE_NAME.value: 'comparison file',
    ConfigOptions.RETURN_FILE.value: False,
    ConfigOptions.VERBOSE.value: False
}


valid_options = {
    ConfigOptions.SBOL_COMPLIANT_URIS.value: {True, False},
    ConfigOptions.SBOL_TYPED_URIS.value: {True, False},
    ConfigOptions.SERIALIZATION_FORMAT.value: {'sbol', 'rdfxml',
                                               'json', 'ntriples'},
    ConfigOptions.VALIDATE.value: {True, False},
    ConfigOptions.LANGUAGE.value: {'SBOL2', 'FASTA', 'GenBank'},
    ConfigOptions.TEST_EQUALITY.value: {True, False},
    ConfigOptions.CHECK_URI_COMPLIANCE.value: {True, False},
    ConfigOptions.CHECK_COMPLETENESS.value: {True, False},
    ConfigOptions.CHECK_BEST_PRACTICES.value: {True, False},
    ConfigOptions.FAIL_ON_FIRST_ERROR.value: {True, False},
    ConfigOptions.PROVIDE_DETAILED_STACK_TRACE.value: {True, False},
    ConfigOptions.INSERT_TYPE.value: {True, False},
    ConfigOptions.RETURN_FILE.value: {True, False},
    ConfigOptions.VERBOSE.value: {True, False}
}


extension_namespaces = {}
# The authoritative namespace for the Document.
# Setting the home namespace is like signing a piece of paper.
home = None
# Flag indicating whether an object's type is included in SBOL-compliant URIs.
SBOLCompliantTypes = 1
catch_exceptions = 0
file_format = 'rdfxml'


class Config:
    """A class which contains global configuration variables
    for the libSBOL environment.

    Configuration variables are accessed
    through the setOption and getOption methods.
    """

    # Dictionary which holds builder functions for SBOL objects. The key is the
    # RDFtype of the object which is parsed from the SBOL file. The dictionary is
    # populated with SBOL core classes in document.py.  The dictionary also supports
    # extension classes with RDFtypes not part of the core specification
    SBOL_DATA_MODEL_REGISTER = {}

    @staticmethod
    def register_extension_class(builder, type_uri):
        """Register an extension class and its namespace, so custom data
        can be embedded into and read from SBOL files.

        :param builder: A no-argument constructor
        :param type_uri: An RDF type URI for the extension class
        """
        Config.SBOL_DATA_MODEL_REGISTER[URIRef(type_uri)] = builder

    @staticmethod
    def setHomespace(ns):
        """Setting the Homespace has several advantages.
        It simplifies object creation and retrieval from Documents.
        In addition, it serves as a way for a user
        to claim ownership of new objects.
        Generally users will want to specify a Homespace
        that corresponds to their organization's web domain.
        :param ns: The namespace to use as the Homespace
        :return: None
        """
        options[ConfigOptions.HOMESPACE.value] = ns

    @staticmethod
    def getHomespace():
        """

        :return: The Homespace (a string representing the default namespace).
        """
        return options[ConfigOptions.HOMESPACE.value]

    @staticmethod
    def hasHomespace():
        """

        :return: True if Homespace is set, False otherwise.
        """
        homespace = ConfigOptions.HOMESPACE.value
        return (homespace in options and
                options[homespace] is not None and
                options[homespace] != "")

    @staticmethod
    def setFileFormat(_file_format):
        """

        :param _file_format: The file format to use.
        :return: None
        """
        # must declare that we're assigning to a global variable
        global file_format
        if _file_format == FileFormats.JSON.value:
            file_format = FileFormats.JSON.value
        elif _file_format == FileFormats.NTRIPLES.value:
            file_format = FileFormats.NTRIPLES.value
        else:
            file_format = FileFormats.RDFXML.value

    @staticmethod
    def getFileFormat():
        """

        :return: The file format.
        """
        return file_format

    @staticmethod
    def setOption(option, val):
        """
        Configure options for libSBOL. Access online validation and conversion.

        | Option                       | Description                                                              | Values          |   # noqa
        | :--------------------------- | :----------------------------------------------------------------------- | :-------------- |
        | homespace                    | Enable validation and conversion requests through the online validator   | http://examples.org |
        | sbol_compliant_uris          | Enables autoconstruction of SBOL-compliant URIs from displayIds          | True or False   |
        | sbol_typed_uris              | Include the SBOL type in SBOL-compliant URIs                             | True or False   |
        | output_format                | File format for serialization                                            | True or False   |
        | validate                     | Enable validation and conversion requests through the online validator   | True or False   |
        | validator_url                | The http request endpoint for validation                                 | A valid URL, set to<br>http://www.async.ece.utah.edu/sbol-validator/endpoint.php by default |
        | language                     | File format for conversion                                               | SBOL2, SBOL1, FASTA, GenBank |
        | test_equality                | Report differences between two files                                     | True or False |
        | check_uri_compliance         | If set to false, URIs in the file will not be checked for compliance<br>with the SBOL specification | True or False |
        | check_completeness           | If set to false, not all referenced objects must be described within<br>the given main_file | True or False |
        | check_best_practices         | If set to true, the file is checked for the best practice rules set<br>in the SBOL specification | True or False |
        | fail_on_first_error          | If set to true, the validator will fail at the first error               | True or False |
        | provide_detailed_stack_trace | If set to true (and failOnFirstError is true) the validator will<br>provide a stack trace for the first validation error | True or False |
        | uri_prefix                   | Required for conversion from FASTA and GenBank to SBOL1 or SBOL2,<br>used to generate URIs  | True or False |
        | version                      | Adds the version to all URIs and to the document                         | A valid Maven version string |
        | return_file                  | Whether or not to return the file contents as a string                   | True or False |
        :param option: The option key
        :param val: The option value (str or bool expected)
        :return: None
        """
        # Convert a config option to its string value
        if isinstance(option, ConfigOptions):
            option = option.value
        # ca-path is deprecated. It is no longer needed in native python
        if option == 'ca-path':
            warnings.warn('ca-path is no longer used', DeprecationWarning)
            return
        if option not in options:
            msg = '{!r} is not a valid configuration option'.format(option)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        if option in valid_options:
            if val in valid_options[option]:
                options[option] = val
            else:
                msg = '{!r} is not a valid value for option {!r}.'.format(val, option)
                msg += ' Valid options are: {!r}.'.format(valid_options[option])
                raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        else:
            # Any argument is valid, eg. uriPrefix
            options[option] = val

    @staticmethod
    def getOption(option):
        """Get current option value for online validation and conversion

        :param option: The option key
        :return: The option value
        """
        # Convert a config option to its string value
        if isinstance(option, ConfigOptions):
            option = option.value
        # ca-path is deprecated. It is no longer needed in native python
        if option == 'ca-path':
            warnings.warn('ca-path is no longer used', DeprecationWarning)
            return ''
        if option in options:
            return options[option]
        else:
            msg = '{!r} is not a valid configuration option'.format(option)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)


# Global methods
def setHomespace(ns):
    Config.setHomespace(ns)


def getHomespace():
    return Config.getHomespace()


def hasHomespace():
    return Config.hasHomespace()


def setFileFormat(file_format):
    Config.setFileFormat(file_format)


def getFileFormat():
    return Config.getFileFormat()


# constructCompliantURI is never invoked. If it ever gets resurrected,
# fix the use of os.sep.
#
# def constructCompliantURI(sbol_type, display_id, version):
#     if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
#         return getHomespace() + os.sep + parseClassName(sbol_type) + os.sep + \
#                display_id + os.sep + version
#     else:
#         return ''


# constructCompliantURI_parentChild is never invoked. If it ever gets
# resurrected, fix the use of os.sep.
#
# def constructCompliantURI_parentChild(parent_type, child_type,
#                                       display_id, version):
#     if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
#         return getHomespace() + os.sep + parseClassName(parent_type) + \
#                os.sep + parseClassName(child_type) + \
#                os.sep + display_id + os.sep + version
#     else:
#         return ''


# randomIdentifier is only invoked by autoconstructURI below, and
# autoconstructURI is never called. Delete this function when
# autoconstructURI is deleted.
#
# def randomIdentifier():
#     # TODO test
#     id = ''
#     for i in range(1, 11):
#         r = random.randint(0, 9)
#         id += str(r)
#         if r % 4 == 0 and r != 16:
#             id += '-'
#     return id


# autoconstructURI is never invoked. If it ever gets resurrected, fix
# the use of os.sep.
#
# def autoconstructURI():
#     if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS) is False \
#             and hasHomespace():
#         return getHomespace() + os.sep + randomIdentifier()
#     elif Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS) is False \
#             and not hasHomespace():
#         raise SBOLError('The autoconstructURI method requires '
#                         'a valid namespace authority. Use setHomespace().',
#                         SBOLErrorCode.SBOL_ERROR_COMPLIANCE)
#     else:
#         raise SBOLError('The autoconstructURI method only works '
#                         'when SBOLCompliance flag is false. '
#                         'Use setOption to disable SBOL-compliant URIs.',
#                         SBOLErrorCode.SBOL_ERROR_COMPLIANCE)


# getCompliantURI is never invoked. If it ever gets resurrected, fix
# the use of os.sep.
#
# def getCompliantURI(uri_prefix, sbol_class_name, display_id, version):
#     return uri_prefix + os.sep + sbol_class_name + \
#            os.sep + display_id + os.sep + version


def parseClassName(uri):
    if '#' in uri:
        return uri[uri.rindex('#')+1:]
    elif '/' in uri:
        return uri[uri.rindex('/')+1:]
    else:
        return ''


def parseNamespace(uri):
    if '#' in uri:
        return uri[:uri.rindex('#')]
    elif '/' in uri:
        return uri[:uri.rindex('/')]
    else:
        return ''


def parseURLDomain(url):
    protocol, path = url.split('://')
    path = path.split('/')[0]
    url = '://'.join([protocol, path])
    return url


def parsePropertyName(uri):
    return parseClassName(uri)


def string_equal(str1: Any, str2: Any) -> bool:
    """Converts the two arguments to str and compares them, returning
    the result. This helps when comparing a str with a rdflib.URIRef
    or a rdflib.Literal.
    """
    return str(str1) == str(str2)



import collections.abc
import logging
import os
import posixpath
import time
from typing import Any, Dict, Mapping, Union
import warnings

from deprecated import deprecated
from rdflib import URIRef

from . import SBOL2Serialize
from . import validation
from .attachment import Attachment
from .collection import Collection
from .combinatorialderivation import CombinatorialDerivation
from .component import Component, FunctionalComponent
from .componentdefinition import ComponentDefinition
from .config import ConfigOptions
from .config import Config
from .config import parseClassName
from . import config
from .config import parsePropertyName
from .constants import *
from .dbtl import Analysis, Build, Design, SampleRoster, Test
from .experiment import Experiment, ExperimentalData
from .identified import Identified
from .implementation import Implementation
from .interaction import Interaction
from .location import Location, Range, Cut, GenericLocation
from .mapsto import MapsTo
from .measurement import Measurement
from .model import Model
from .module import Module
from .moduledefinition import ModuleDefinition
from .object import SBOLObject
from .participation import Participation
from .property import OwnedObject, URIProperty
from .provo import Plan, Activity, Agent, Usage, Association
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from .sequence import Sequence
from .sequenceannotation import SequenceAnnotation
from .sequenceconstraint import SequenceConstraint
from .toplevel import TopLevel
from .uridict import URIDict

import requests

Config.SBOL_DATA_MODEL_REGISTER = {
    URIRef(UNDEFINED): SBOLObject,
    URIRef(SBOL_IDENTIFIED): Identified,
    URIRef(SBOL_COMPONENT_DEFINITION): ComponentDefinition,
    URIRef(SBOL_SEQUENCE_ANNOTATION): SequenceAnnotation,
    URIRef(SBOL_SEQUENCE): Sequence,
    URIRef(SBOL_COMPONENT): Component,
    URIRef(SBOL_FUNCTIONAL_COMPONENT): FunctionalComponent,
    URIRef(SBOL_MODULE_DEFINITION): ModuleDefinition,
    URIRef(SBOL_MODULE): Module,
    URIRef(SBOL_INTERACTION): Interaction,
    URIRef(SBOL_PARTICIPATION): Participation,
    URIRef(SBOL_MODEL): Model,
    URIRef(SBOL_SEQUENCE_CONSTRAINT): SequenceConstraint,
    URIRef(SBOL_RANGE): Range,
    URIRef(SBOL_MAPS_TO): MapsTo,
    URIRef(SBOL_CUT): Cut,
    URIRef(SBOL_COLLECTION): Collection,
    URIRef(SBOL_LOCATION): Location,
    URIRef(SBOL_GENERIC_LOCATION): GenericLocation,
    URIRef(PROVO_PLAN): Plan,
    URIRef(PROVO_ACTIVITY): Activity,
    URIRef(PROVO_AGENT): Agent,
    URIRef(PROVO_USAGE): Usage,
    URIRef(PROVO_ASSOCIATION): Association,
    URIRef(SBOL_ATTACHMENT): Attachment,
    URIRef(SBOL_COMBINATORIAL_DERIVATION): CombinatorialDerivation,
    URIRef(SBOL_IMPLEMENTATION): Implementation,
    URIRef(SYSBIO_DESIGN): Design,
    URIRef(SYSBIO_ANALYSIS): Analysis,
    URIRef(SYSBIO_SAMPLE_ROSTER): SampleRoster,
    URIRef(SBOL_EXPERIMENT): Experiment,
    URIRef(SBOL_EXPERIMENTAL_DATA): ExperimentalData,
    URIRef(OM_MEASURE): Measurement,
}


class Document(Identified):
    """
    The Document is a container for all SBOL data objects.

    In a previous era, engineers might sit at a drafting board
    and draft a design by hand.
    The engineer's drafting sheet in LibSBOL is called a Document.
    The Document serves as a container,
    initially empty, for SBOL data objects.
    All file I/O operations are performed on the Document
    to populate it with SBOL objects representing design elements.
    """

    def __init__(self, filename=None):
        """
        Construct a document.

        :param filename: (optional) a file to initialize the Document.
        """
        super().__init__(SBOL_DOCUMENT, URIRef(""), VERSION_STRING)

        # A RDFLib representation of the triples.
        # Initialized when parsing a graph.
        # Updated when writing a graph.
        self.graph = rdflib.Graph()
        # The Document's register of objects
        self.objectCache = {}  # Needed?

        # The keys in SBOLObjects are URIRefs because our internal
        # representations use rdflib types.
        # self.SBOLObjects: Dict[rdflib.URIRef, SBOLObject] = {}
        self.SBOLObjects = URIDict()

        self._namespaces = {}
        self.resource_namespaces = set()
        self.designs = OwnedObject(self, SYSBIO_DESIGN, Design,
                                   '0', '*', [validation.libsbol_rule_11])
        self.builds = OwnedObject(self, SYSBIO_BUILD, Build,
                                  '0', '*', [validation.libsbol_rule_12])
        self.tests = OwnedObject(self, SYSBIO_TEST, Test,
                                 '0', '*', [validation.libsbol_rule_13])
        self.analyses = OwnedObject(self, SYSBIO_ANALYSIS, Analysis,
                                    '0', '*', [validation.libsbol_rule_14])
        self.componentDefinitions = OwnedObject(self,
                                                SBOL_COMPONENT_DEFINITION,
                                                ComponentDefinition,
                                                '0', '*', None)
        self.moduleDefinitions = OwnedObject(self,
                                             SBOL_MODULE_DEFINITION,
                                             ModuleDefinition,
                                             '0', '*', None)
        self.models = OwnedObject(self, SBOL_MODEL, Model,
                                  '0', '*', None)
        self.sequences = OwnedObject(self, SBOL_SEQUENCE, Sequence,
                                     '0', '*', None)
        self.collections = OwnedObject(self, SBOL_COLLECTION, Collection,
                                       '0', '*', None)
        self.activities = OwnedObject(self, PROVO_ACTIVITY, Activity,
                                      '0', '*', None)
        self.plans = OwnedObject(self, PROVO_PLAN, Plan,
                                 '0', '*', None)
        self.agents = OwnedObject(self, PROVO_AGENT, Agent,
                                  '0', '*', None)
        self.attachments = OwnedObject(self, SBOL_ATTACHMENT, Attachment,
                                       '0', '*', None)
        self.combinatorialderivations = OwnedObject(self,
                                                    SBOL_COMBINATORIAL_DERIVATION,
                                                    CombinatorialDerivation,
                                                    '0', '*', None)
        self.implementations = OwnedObject(self, SBOL_IMPLEMENTATION,
                                           Implementation,
                                           '0', '*', None)
        self.sampleRosters = OwnedObject(self, SYSBIO_SAMPLE_ROSTER,
                                         SampleRoster,
                                         '0', '*', [validation.libsbol_rule_16])
        self.experiments = OwnedObject(self, SBOL_EXPERIMENT, Experiment,
                                       '0', '*', None)
        self.experimentalData = OwnedObject(self, SBOL_EXPERIMENTAL_DATA,
                                            ExperimentalData,
                                            '0', '*', None)
        self.citations = URIProperty(self, PURL_URI + "bibliographicCitation",
                                     '0', '*', None)
        self.keywords = URIProperty(self, PURL_URI + "elements/1.1/subject",
                                    '0', '*', None)
        # I am my own document
        self.doc = self
        if filename is not None:
            self.read(filename)

    def compare(self, other):
        # Let the super class do the bulk of the comparison. Super
        # compares owned objects and properties.
        if not super().compare(other):
            return False
        if self._namespaces != other._namespaces:
            return False
        return True

    def addCitation(self, new_citation):
        val = self.citations
        val.append(new_citation)
        self.citations = val

    def removeCitation(self, index=0):
        val = self.citations
        del val[index]
        self.citations = val

    def addKeyword(self, new_keyword):
        val = self.keywords
        val.append(new_keyword)
        self.keywords = val

    def removeKeyword(self, index=0):
        val = self.keywords
        del val[index]
        self.keywords = val

    def add(self, sbol_obj):
        """
        Register an object in the Document.

        :param sbol_obj: The SBOL object(s) you want to serialize.
            Either a single object or a list of objects.
        :return: None
        """
        # Check for uniqueness of URI
        identity_uri = sbol_obj.identity
        if identity_uri in self.SBOLObjects:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE,
                            'Cannot add ' + sbol_obj.identity +
                            ' to Document. An object with this identity '
                            'is already contained in the Document')
        else:
            # If TopLevel add to Document.
            if sbol_obj.is_top_level():
                self.SBOLObjects[identity_uri] = sbol_obj
            type_uri = rdflib.URIRef(sbol_obj.getTypeURI())
            if type_uri in self.owned_objects:
                sbol_obj.parent = self  # Set back-pointer to parent object
                # Add the object to the Document's property store,
                # eg. componentDefinitions, moduleDefinitions, etc.
                self.owned_objects[type_uri].append(sbol_obj)
            sbol_obj.doc = self
            # Notify the object that it has been added
            sbol_obj._added_to_document(self)
            # Recurse into child objects and set their back-pointer to this Document
            for key, obj_store in sbol_obj.owned_objects.items():
                for child_obj in obj_store:
                    if child_obj.doc != self:
                        self.add(child_obj)

    def add_list(self, sbol_objs):
        for obj in sbol_objs:
            self.add(obj)

    def addNamespace(self, namespace, prefix):
        """Add a new namespace to the Document.

        :param namespace: The namespace, eg. http://sbols.org/v2#
        :param prefix: The namespace prefix, eg. sbol
        :return:
        """
        if namespace[-1] != '#' and namespace[-1] != '/':
            raise ValueError('Invalid namespace. Namespace must end with # or /')
        namespace = rdflib.term.URIRef(namespace)
        # Overwrite the prefix, if the namespace already has a conflicting one
        inv_namespace_map = {ns: p for p, ns in self._namespaces.items()}
        inv_namespace_map[namespace] = prefix
        self._namespaces = {p: ns for ns, p in inv_namespace_map.items()}

    def addComponentDefinition(self, sbol_obj):
        """
        Convenience method for adding a component definition.

        :param sbol_obj: component definition
        :return: None
        """
        if isinstance(sbol_obj, list):
            for obj in sbol_obj:
                self.add(obj)
        else:
            self.add(sbol_obj)

    def addModuleDefinition(self, sbol_obj):
        """
        Convenience method for adding a module definition.

        :param sbol_obj: module definition
        :return: None
        """
        if isinstance(sbol_obj, collections.abc.Iterable):
            for obj in sbol_obj:
                self.add(obj)
        else:
            self.add(sbol_obj)

    def addCollection(self, sbol_obj):
        """
        Convenience method for adding a collection.

        :param sbol_obj: a Collection
        :return: None
        """
        if isinstance(sbol_obj, collections.abc.Iterable):
            for obj in sbol_obj:
                self.add(obj)
        else:
            self.add(sbol_obj)

    def addSequence(self, sbol_obj):
        """
        Convenience method for adding a sequence.

        :param sbol_obj: sequence
        :return: None
        """
        self.add(sbol_obj)

    def addModel(self, sbol_obj):
        """
        Convenience method for adding a model.

        :param sbol_obj: model
        :return: None
        """
        self.add(sbol_obj)

    def addImplementation(self, implementation):
        """ Convenience method for adding an implementation.
        """
        if isinstance(implementation, collections.abc.Iterable):
            for impl in implementation:
                self.add(impl)
        else:
            self.add(implementation)

    def addAttachment(self, attachment):
        """Add an attachment to this document.
        """
        if isinstance(attachment, collections.abc.Iterable):
            for a in attachment:
                self.add(a)
        else:
            self.add(attachment)

    def create(self, uri):
        """
        Creates another SBOL object derived from TopLevel
        and adds it to the Document.
        NOTE: originally from ReferencedObject
        :param uri: In "open world" mode, this is a full URI
        and the same as the returned URI.
        If the default namespace for libSBOL has been configured,
        then this argument should simply be a local identifier.
        If SBOL-compliance is enabled,
        this argument should be the intended displayId of the new object.
        A full URI is automatically generated and returned.
        :return: The full URI of the created object
        """
        raise NotImplementedError("Not yet implemented")

    def get(self, uri):
        """ Finds the SBOL object with the given URI.

        :param uri: URI of the extension object
        :type uri: str
        :return: The matching SBOLObject
        :rtype: SBOLObject
        :raises: SBOLError if the given uri is not found
        """
        try:
            return self.SBOLObjects[uri]
        except KeyError:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            f'Object {uri} was not found')

    def getAll(self):
        """
        Retrieve a list of objects from the Document.

        :return: A list of objects from the Document.
        """
        raise NotImplementedError("Not yet implemented")

    def getComponentDefinition(self, uri):
        # NOTE: I couldn't find this in the original libSBOL source,
        # but they are heavily used in all the unit tests.
        return self.componentDefinitions.get(uri)

    def getModuleDefinition(self, uri):
        return self.moduleDefinitions.get(uri)

    def getSequence(self, uri):
        return self.sequences.get(uri)

    def getModel(self, uri):
        return self.models.get(uri)

    def getImplementation(self, uri):
        return self.implementations.get(uri)

    def getCollection(self, uri: str) -> Collection:
        return self.collections.get(uri)

    # File I/O #
    def write(self, filename):
        """
        Serialize all objects in this Document to an RDF/XML file.

        :param filename: The full name of the file you want to write
        (including file extension).
        :return: A string with the validation results,
        or empty string if validation is disabled.
        """
        self.doc_serialize_rdf2xml(filename)
        # Optionally validate
        result = 'Validation disabled. To enable use of the online validation tool, use'
        result += ' Config.setOption(ConfigOptions.VALIDATE, True)'
        if Config.getOption(ConfigOptions.VALIDATE):
            t_start = time.time()
            result = self.validate()
            if Config.getOption(ConfigOptions.VERBOSE):
                t_end = time.time()
                print(f'Validation request took {t_end - t_start} seconds')
        return result

    def read(self, filename):
        """
        Read an RDF/XML file and attach the SBOL objects to this Document.

        Existing contents of the Document will be wiped.
        :param filename: The full name of the file you want to read
        (including file extension).
        :return: None
        """
        self.clear()
        self.append(filename, overwrite=False)

    def readString(self, sbol_str):
        """Read an RDF/XML string and attach the SBOL objects to
        this Document.

        Existing contents of the Document will be wiped.

        :param sbol_str: A string formatted in SBOL.
        :return: None
        """
        self.clear()
        self.appendString(sbol_str, overwrite=False)

    def writeString(self):
        """
        Convert data objects in this Document into textual SBOL.

        :return: A string representation of the objects in this Document.
        """
        # Save any changes we've made to the graph.
        self.update_graph()
        # Write graph to string
        rdf = SBOL2Serialize.serialize_sboll2(self.graph).decode('utf-8')
        return rdf

    def append(self, filename, overwrite: bool = False):
        """
        Read an RDF/XML file and attach the SBOL objects to this Document.

        New objects will be added to the existing contents of the Document.
        :param filename: The full name of the file you want to read
        (including file extension).
        :param overwrite: Boolean indicating whether to overwrite existing objects
        :return: None
        """
        new_graph = rdflib.Graph()
        new_graph.parse(filename, format='application/rdf+xml')
        self._append_graph(new_graph, overwrite)

    def appendString(self, sbol_str: str, overwrite: bool = False):
        """
        Read an RDF/XML document from a string and attach the SBOL
        objects to this Document.

        New objects will be added to the existing contents of the Document.
        :param sbol_str: A string of RDF/XML
        :param overwrite: Boolean indicating whether to overwrite existing objects
        :return: None
        """
        # ------------------------------------------------------------
        # Load the new data into a graph
        new_graph = rdflib.Graph()
        new_graph.parse(data=sbol_str, format='application/rdf+xml')
        self._append_graph(new_graph, overwrite)

    def _append_graph(self, new_graph: rdflib.Graph, overwrite: bool):
        # Gather all the objects that will be overwritten, stopping
        # if the user says not to overwrite. If we clear as we go we lose
        # the ability to find objects within objects. So gather the list
        # here, and clear them as a second pass.
        objects_to_clear = []
        identities = (s for s, _, _ in
                      new_graph.triples((None, rdflib.RDF.type, None)))
        objects = (self.find(identity) for identity in identities)
        objects_to_clear = [obj for obj in objects if obj is not None]
        if overwrite is False and objects_to_clear:
            msg = objects_to_clear[0].identity
            count = len(objects_to_clear)
            if count > 1:
                msg += f' and {count - 1} others'
            msg += ' would require overwriting'
            raise SBOLError(SBOLErrorCode.DUPLICATE_URI_ERROR, msg)
        # Clear out the internal stores of the objects_to_clear so that
        # they will be overwritten. Keep the identity property because it
        # does not get restored by the graph parsing. Keep all the keys in
        # the internal stores so owned objects end up in the right place.
        for existing_object in objects_to_clear:
            if existing_object.identity not in self.SBOLObjects:
                self.SBOLObjects[existing_object.identity] = existing_object
            # Now clear the object. It will get reconstituted below by parse_all
            for k in existing_object.properties:
                # Clear out everything but the identity
                if k == SBOL_IDENTITY:
                    continue
                existing_object.properties[k] = []
            for k in existing_object.owned_objects:
                existing_object.owned_objects[k] = []
        # Make the new graph be the graph we parse
        self.graph = new_graph
        # Load the new graph into the existing document
        self.parse_all()

    def parse_all(self):
        # Parse namespaces
        self.logger.debug("*** Reading in namespaces (graph): ")
        for ns in self.graph.namespaces():
            self.logger.debug(ns)
            self._namespaces[ns[0]] = ns[1]
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("*** Internal namespaces data structure: ")
            for ns in self._namespaces:
                self.logger.debug(ns)
        # Instantiate all objects with an RDF type
        for s, _, o in self.graph.triples((None, rdflib.RDF.type, None)):
            self.parse_objects_inner(s, o)
        # Find the graph base uri.  This is the location of the sbol
        # file, and begins with the "file://" scheme.  Any URI in the
        # file without a scheme will appear relative to this URI, after
        # the file is parsed.  Therefore, if the any URI property value
        # begins with the graph base uri, the base part of the URI is removed.
        graphBaseURIStr = "file://" + os.getcwd()  # Not sure if this is correct...
        # Remove the filename from the path
        pos = graphBaseURIStr.rfind('/')
        if pos != -1:
            pos += 1
        rdf_type = rdflib.RDF.type
        for result_s, result_p, result_o in self.graph:
            # Look for properties
            if result_p != rdf_type:
                obj = result_o
                if isinstance(result_o, URIRef) and pos != -1:
                    if obj[:pos] == graphBaseURIStr:
                        # This was a URI without a scheme.  Remove URI base
                        obj = URIRef(obj[pos:])
                self.parse_properties_inner(result_s, result_p, obj)

        # Remove objects from SBOLObjects if they are not TopLevel AND
        # they have a parent object.
        #
        # Note: use a list of the keys so that we can modify the dict
        # while we iterate.
        for k in list(self.SBOLObjects.keys()):
            so = self.SBOLObjects[k]
            if isinstance(so, TopLevel):
                continue
            if so.parent:
                # Not TopLevel and already has parent, remove from
                # SBOLObjects
                del self.SBOLObjects[k]
                continue
            self.logger.debug('Orphan %r', so)

        # Handle the annotation objects
        self.parse_annotation_objects()
        # Dress document
        self.dress_document()

    def parse_objects_inner(self, subject, obj):
        # Construct the top-level object if we haven't already done so
        # and its type is something we know about.
        if subject not in self.SBOLObjects and obj in Config.SBOL_DATA_MODEL_REGISTER:
            # Call constructor for the appropriate SBOLObject
            new_obj = Config.SBOL_DATA_MODEL_REGISTER[obj]()
            if isinstance(new_obj, Identified):
                # Clear out the version. it will get set later
                new_obj.version = ''
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug("New object type: " + str(type(new_obj)))
                self.logger.debug("New object attrs: " + str(vars(new_obj)))
            # Wipe default property values passed from default
            # constructor. New property values will be added as properties
            # are parsed from the input file
            for prop_name, values in new_obj.properties.items():
                values.clear()
            new_obj.identity = subject
            # Update document
            self.SBOLObjects[new_obj.identity] = new_obj
            new_obj.doc = self
            # For now, set the parent to the Document.
            # This may get overwritten later for child objects.
            new_obj.parent = self
            # If the new object is TopLevel,
            # add to the Document's property store
            if new_obj.is_top_level():
                if new_obj.rdf_type not in self.owned_objects:
                    self.owned_objects[new_obj.rdf_type] = []
                self.owned_objects[new_obj.rdf_type].append(new_obj)
        elif (subject not in self.SBOLObjects
              and obj not in Config.SBOL_DATA_MODEL_REGISTER):
            # Generic TopLevels
            new_obj = SBOLObject()
            new_obj.identity = subject
            new_obj.rdf_type = obj
            self.SBOLObjects[new_obj.identity] = new_obj
            new_obj.doc = self

    def parse_properties_inner(self, subject, predicate, obj):
        self.logger.debug('Adding (%r, %r, %r)', subject, predicate, obj)
        found = predicate.rfind('#')
        if found == -1:
            found = predicate.rfind('/')
        if found != -1:
            # Checks if the object's property already exists
            if subject in self.SBOLObjects:
                parent = self.SBOLObjects[subject]
                # Decide if this triple corresponds to a simple property,
                # a list property, an owned property or a referenced property
                if predicate in parent.properties:
                    # triple is a property
                    if obj not in parent.properties[predicate]:
                        parent.properties[predicate].append(obj)
                elif predicate in parent.owned_objects:
                    # triple is an owned object
                    owned_obj = self.SBOLObjects[obj]
                    if owned_obj is not None:
                        if owned_obj not in parent.owned_objects[predicate]:
                            parent.owned_objects[predicate].append(owned_obj)
                            owned_obj.parent = parent
                else:
                    # Extension data
                    parent.properties[predicate] = [obj]
            else:
                msg = 'Subject {} ({}) not found in my SBOLObjects'
                msg = msg.format(subject, type(subject))
                self.logger.debug(msg)

    def find_reference(self, uri):
        """Find objects that reference the given URI. Returns a list of
        objects. The list will be empty if no references were found.

        """
        references = []
        for obj in self.SBOLObjects.values():
            references.extend(obj.find_reference(uri))
        return references

    def parse_annotation_objects(self):
        """Parse leftover objects from reading and link them up where they
        belong. These are usually extension-type objects.

        """
        for uri, obj in self.SBOLObjects.items():
            self.logger.debug('Possible annotation object %s', obj.identity)
        annotation_objects = [obj for obj in self.SBOLObjects.values()
                              if not isinstance(obj, TopLevel)]
        for ao in annotation_objects:
            self.logger.debug('Annotation object: %s', ao.identity)
            if SBOL_PERSISTENT_IDENTITY in ao.properties:
                # Copy to a new TopLevel object
                tl = TopLevel(type_uri=ao.type, version=None)
                for k, v in ao.properties.items():
                    tl.properties[k] = v
                for k, v in ao.owned_objects.items():
                    tl.owned_objects[k] = v
                tl.doc = self
                tl_identity_uri = rdflib.URIRef(tl.identity)
                self.SBOLObjects[tl_identity_uri] = tl
            else:
                # Determine the RDF type of the member property that
                # contains this kind of annotation object
                ao_identity_uri = rdflib.URIRef(ao.identity)
                ns = config.parseNamespace(ao.type)
                self.logger.debug('anno ns = %r', ns)
                class_name = config.parseClassName(ao.type)
                self.logger.debug('anno class name = %r', class_name)
                # lowercase the first character
                property_name = class_name[0].lower() + class_name[1:]
                self.logger.debug('anno property name = %r', property_name)
                property_uri = rdflib.URIRef(posixpath.join(ns, property_name))
                self.logger.debug('anno property uri = %r', property_uri)
                matches = self.find_reference(ao_identity_uri)
                self.logger.debug('Found %d references', len(matches))
                matches = [m for m in matches if property_uri in m.properties]
                self.logger.debug('Found %d good references', len(matches))
                if len(matches) > 1:
                    msg = 'Invalid custom annotation object in SBOL document'
                    raise SBOLError(SBOLErrorCode.SBOL_ERROR_SERIALIZATION, msg)
                if len(matches) == 1:
                    match = matches[0]
                    if property_uri not in match.owned_objects:
                        match.owned_objects[property_uri] = []
                    match.owned_objects[property_uri].append(ao)
                    ao.parent = match
                    # Clean up the property
                    self.logger.debug('match[%r] = %r', property_uri,
                                      match.properties[property_uri])
                    self.logger.debug('ao.identity = %r', ao.identity)
                    match.properties[property_uri].remove(ao_identity_uri)
                    if len(match.properties[property_uri]) == 0:
                        # Remove the empty property list
                        del match.properties[property_uri]
                    # Remove this annotation object from the list of
                    # SBOLObjects
                    ao_identity_uri = rdflib.URIRef(ao.identity)
                    del self.SBOLObjects[ao_identity_uri]

    def infer_resource_namespaces(self):
        for obj in self.SBOLObjects.values():
            if not isinstance(obj, Identified):
                continue
            if not (obj.persistentIdentity and obj.displayId and obj.version):
                continue
            # if object identity ends with compliant suffix, extract the
            # start as a namespace
            uri = obj.identity
            compliant_suffix = posixpath.join(posixpath.sep, obj.displayId,
                                              obj.version)
            if uri.endswith(compliant_suffix):
                self.resource_namespaces.add(uri[0:-len(compliant_suffix)])
                continue
            typed_suffix = posixpath.join(posixpath.sep, parseClassName(obj.rdf_type),
                                          obj.displayId, obj.version)
            if uri.endswith(typed_suffix):
                self.resource_namespaces.add(uri[0:-len(typed_suffix)])

    def dress_document(self):
        self.infer_resource_namespaces()
        # There is a lot more that is done in libSBOL Document::dress_document()
        # TODO: do more of that here

    def convert_ntriples_encoding_to_ascii(self, s):
        s.replace("\\\"", "\"")
        s.replace("\\\\", "\\")
        return s

    def request_comparison(self, diff_file):
        """
        Perform comparison on Documents using the online validation tool.

        This is for cross-validation of SBOL documents with libSBOLj.
        Document comparison can also be performed using the built-in compare method.
        :param diff_file:
        :return: The comparison results
        """
        raise NotImplementedError("Not yet implemented")

    def clear(self, clear_graph=True):
        """
        Delete all properties and objects in the Document.

        :return: None
        """
        # Properties to keep, which don't make sense to clear
        keepers = [SBOL_VERSION]
        self.SBOLObjects.clear()
        for name, value in self.properties.items():
            if name in keepers:
                # Do not erase properties on the keepers list
                continue
            value.clear()
        for object_store in self.owned_objects.values():
            object_store.clear()
        self._namespaces.clear()
        self.resource_namespaces.clear()
        if clear_graph:
            self.graph = rdflib.Graph()  # create a new graph

    def query_repository(self, command):
        """

        :param command:
        :return: str
        """
        # TODO better docstring
        raise NotImplementedError("Not yet implemented")

    def search_metadata(self, role, type, name, collection):
        """

        :param role:
        :param type:
        :param name:
        :param collection:
        :return: str
        """
        # TODO better docstring
        raise NotImplementedError("Not yet implemented")

    def doc_serialize_rdf2xml(self, outfile):
        """
        Serialize RDF XML.
        :param outfile: output file
        :return: None
        """
        self.update_graph()
        rdf = SBOL2Serialize.serialize_sboll2(self.graph).decode('utf-8')
        self.logger.debug("RDF: " + rdf)
        self.logger.debug("TYPE: " + str(type(rdf)))
        with open(outfile, 'w') as out:
            out.write(rdf)
            out.flush()

    def update_graph(self):
        """
        Update the RDF triples representation of data.
        :return:
        """
        self.graph = rdflib.Graph()
        for prefix, ns in self._namespaces.items():
            self.graph.bind(prefix, ns)
        # ASSUMPTION: Document does not have properties. Is this a valid assumption?
        for obj in self.SBOLObjects.values():
            obj.build_graph(self.graph)
        if self.logger.isEnabledFor(logging.DEBUG):
            for s, p, o in self.graph:
                self.logger.debug('Graph contains: %r', (s, p, o))

    def validate(self):
        """
        Run validation on this Document via the online validation tool.

        :return: A string containing a message with the validation results
        :rtype: str
        """
        response = validate(self, config.options)
        if response['valid']:
            result = "Valid."
        else:
            result = "Invalid."
        errors = ' '.join(response['errors'])
        if errors:
            result = ' '.join([result, errors])
        return result

    def size(self):
        """
        Get the total number of objects in the Document,
        including SBOL core object and custom annotation objects.

        :return: The total number of objects in the Document.
        """
        return len(self.SBOLObjects)

    def __len__(self):
        """
        Get the total number of objects in the Document,
        including SBOL core object and custom annotation objects.

        (Returns the same thing as size())

        :return: The total number of objects in the Document.
        """
        return self.size()

    def __bool__(self):
        # Ignore list semantics for boolean conversion.
        # A document is True if it exists.
        return True

    def __str__(self):
        """
        Produce a string representation of the Document.

        :return: A string representation of the Document.
        """
        return self.summary()

    # def __iter__(self):
    #     self.current_obj = 0
    #     self.owned_objects_list = []
    #     for objlist in self.owned_objects.values():
    #         for obj in objlist:
    #             self.owned_objects_list.append(obj)
    #     return self
    #

    def __iter__(self):
        self.current_obj = 0
        self.owned_objects_list = []
        for obj in self.SBOLObjects.values():
            self.owned_objects_list.append(obj)
        return self

    def __next__(self):
        if self.current_obj > len(self.owned_objects_list)-1:
            raise StopIteration
        else:
            ret = self.owned_objects_list[self.current_obj]
            self.current_obj += 1
            return ret

    def cacheObjectsDocument(self):
        # TODO docstring
        raise NotImplementedError("Not yet implemented")

    def referenceNamespace(self, uri):
        """Replaces the namespace with a reference
        and removes the default namespace, shortening the URI.
        :param uri:
        :return: str
        """
        if self._default_namespace is not None and \
                len(self._default_namespace) > 0:
            if self._default_namespace in uri:
                uri.replace(self._default_namespace, '')
                return uri
        for abbrev, ns in self._namespaces.items():
            if ns in uri:
                # Assume only one namespace per URI
                uri.replace(ns, abbrev)
                return uri

    def summary(self):
        """
        Produce a string representation of the Document.

        :return: A string representation of the Document.
        """
        summary = ''
        col_size = 30
        total_core_objects = 0
        for rdf_type, obj_store in self.owned_objects.items():
            property_name = parsePropertyName(rdf_type)
            obj_count = len(obj_store)
            total_core_objects += obj_count
            summary += property_name
            summary += '.' * (col_size-len(property_name))
            summary += str(obj_count) + '\n'
        summary += 'Annotation Objects'
        summary += '.' * (col_size-18)
        summary += str(self.size() - total_core_objects) + '\n'
        summary += '---\n'
        summary += 'Total: '
        summary += '.' * (col_size-5)
        summary += str(self.size()) + '\n'
        return summary

    # TODO Port iterator, which loops over top-level items of Document

    def find(self, uri):
        """
        Search recursively for an SBOLObject in this Document
        that matches the uri.

        :param uri: The identity of the object to search for.
        :return: A pointer to the SBOLObject,
        or NULL if an object with this identity doesn't exist.
        """
        for obj in self.SBOLObjects.values():
            match = obj.find(uri)
            if match is not None:
                return match
        return None

    def getTypeURI(self):
        return URIRef(SBOL_DOCUMENT)

    def getTopLevel(self, uri):
        # Ensure it's a URI Ref
        uri = rdflib.URIRef(uri)
        if uri not in self.SBOLObjects:
            msg = 'Top level object {} is not in document'
            msg = msg.format(uri)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND, msg)
        sbol_obj = self.SBOLObjects[uri]
        # Verify object is top level
        if sbol_obj.is_top_level():
            return sbol_obj
        # Not top level, raise error
        msg = '{} is not a top level object'
        msg = msg.format(uri)
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)

    def copy(self, target_namespace=None, target_doc=None, version=None):
        # This enables the user to use the pattern doc2 = doc.copy() to clone a Document.
        # SWIG pySBOL assumes the user does NOT want to increment the Document's version
        # when creating a clone. (Whether or not these are the right semantics, that's
        # how it is implemented.) By passing an explicit version parameter to super's
        # copy method, we short-circuit its default behavior to auto-increment version.
        if version is None:
            version = self.version
        return super().copy(target_doc, target_namespace, version)

    def exportToFormat(self, language: str, output_path: str):
        # Copy the global config options. Shallow copy is ok because values
        # are either bool or str.
        options = config.options.copy()
        options[ConfigOptions.LANGUAGE.value] = language
        # We always want the return file
        options[ConfigOptions.RETURN_FILE.value] = True
        response = validate(self, options)

        # What should we be expecting from the validator?
        # Can we tell if there was an error?
        # Is it if there are any errors? Or if 'Conversion failed.' is one of the errors?
        # Or if response['result'] is not empty?
        if response['errors'][0]:
            msg = ' '.join(response['errors'])
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        if not response['result']:
            msg = 'Validator returned no content'
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)
        # write the result to the desired output path
        with open(output_path, 'w') as fp:
            fp.write(response['result'])

    def convert(self, language, output_path):
        warnings.warn('Document.convert is now Document.exportToFormat',
                      DeprecationWarning)
        self.exportToFormat(language, output_path)

    def importFromFormat(self, input_path: str, overwrite=False):
        """Import the specified file into this document.
        """
        # The C++ version of this function takes a language argument
        # and does not use it. There is no need for a language
        # argument based on the validator API.
        # Copy the global config options. Shallow copy is ok because values
        # are either bool or str.
        options = config.options.copy()
        # We want an SBOL2 file back
        options[ConfigOptions.LANGUAGE.value] = 'SBOL2'
        options[ConfigOptions.URI_PREFIX.value] = Config.getHomespace()

        json_request = _make_validation_request(options)
        # We always want the return file.
        json_request[ConfigOptions.RETURN_FILE.value] = True

        # get the input file data to send
        with open(input_path, 'r') as infile:
            contents = infile.read()
        json_request['main_file'] = contents

        import json
        with open('genbank2sbol.json', 'w') as outjson:
            json.dump(json_request, outjson, indent=4)

        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'charsets': 'utf-8'
        }

        validator_url = options[ConfigOptions.VALIDATOR_URL.value]

        # Send the request to the online validation tool
        response = requests.post(validator_url,
                                 json=json_request,
                                 headers=headers)
        if response:
            response = response.json()
        else:
            msg = 'Validation failure. HTTP post request failed with code {}: {}'
            msg = msg.format(response.status_code, response.content)
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)
        if response['valid']:
            self.appendString(response['result'], overwrite)
        else:
            msg = ' '.join(response['errors'])
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_INVALID_ARGUMENT, msg)

    @deprecated(reason="Use Document.get() instead")
    def getExtensionObject(self, uri: str) -> SBOLObject:
        """*Deprecated.* Use Document.get instead.

        :param uri: URI of the extension object
        :type uri: str
        :return: The matching SBOLObject
        :rtype: SBOLObject
        :raises: SBOLError if the given uri is not found
        """
        return self.get(uri)

    @deprecated(reason="Use Document.add() instead")
    def addExtensionObject(self, obj: SBOLObject) -> None:
        """
        *Deprecated.* Use Document.add instead.

        :param obj: An SBOLObject to add to this document
        :type obj: SBOLObject
        :return: None
        """
        # Just do an add
        self.add(obj)


def _make_validation_request(options: Mapping[str, Union[bool, str]]):
    config_options = [
        config.ConfigOptions.CHECK_BEST_PRACTICES.value,
        config.ConfigOptions.CHECK_COMPLETENESS.value,
        config.ConfigOptions.CHECK_URI_COMPLIANCE.value,
        config.ConfigOptions.DIFF_FILE_NAME.value,
        config.ConfigOptions.FAIL_ON_FIRST_ERROR.value,
        config.ConfigOptions.INSERT_TYPE.value,
        config.ConfigOptions.LANGUAGE.value,
        config.ConfigOptions.MAIN_FILE_NAME.value,
        config.ConfigOptions.PROVIDE_DETAILED_STACK_TRACE.value,
        config.ConfigOptions.SUBSET_URI.value,
        config.ConfigOptions.TEST_EQUALITY.value,
        config.ConfigOptions.URI_PREFIX.value,
        config.ConfigOptions.VERSION.value
    ]
    request_options = {}
    for key in config_options:
        request_options[key] = options[key]
    return dict(options=request_options)


def validate(doc: Document, options: Mapping[str, Any]):
    """
    :rtype: Dict[str, Any]
    """
    return_file_key = config.ConfigOptions.RETURN_FILE.value
    validator_key = config.ConfigOptions.VALIDATOR_URL.value
    json_request = _make_validation_request(options)
    # We always want the return file
    json_request[return_file_key] = options[return_file_key]
    json_request['main_file'] = doc.writeString()

    headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'charsets': 'utf-8'
    }

    validator_url = options[validator_key]

    # Send the request to the online validation tool
    response = requests.post(validator_url,
                             json=json_request,
                             headers=headers)
    if response:
        return response.json()
    else:
        msg = 'Validation failure. HTTP post request failed with code {}: {}'
        msg = msg.format(response.status_code, response.content)
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_BAD_HTTP_REQUEST, msg)


igem_assembly_scars = '''<?xml version="1.0" encoding="utf-8"?>
<rdf:RDF xmlns:dc="http://purl.org/dc/elements/1.1/"
xmlns:dcterms="http://purl.org/dc/terms/"
xmlns:gbconv="http://sbols.org/genBankConversion#"
xmlns:genbank="http://www.ncbi.nlm.nih.gov/genbank#"
xmlns:igem="http://wiki.synbiohub.org/wiki/Terms/igem#"
xmlns:ncbi="http://www.ncbi.nlm.nih.gov#"
xmlns:obo="http://purl.obolibrary.org/obo/"
xmlns:om="http://www.ontology-of-units-of-measure.org/resource/om-2/"
xmlns:prov="http://www.w3.org/ns/prov#"
xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
xmlns:sbh="http://wiki.synbiohub.org/wiki/Terms/synbiohub#"
xmlns:sbol="http://sbols.org/v2#"
xmlns:sybio="http://www.sybio.ncl.ac.uk#"
xmlns:synbiohub="http://synbiohub.org#"
xmlns:xsd="http://www.w3.org/2001/XMLSchema#dateTime/">
<sbol:ComponentDefinition rdf:about="https://synbiohub.org/public/igem/BBa_G0000/1">
<dc:creator>Reshma Shetty</dc:creator>
<dcterms:created>2007-07-22T11:00:00Z</dcterms:created>
<dcterms:description>SpeI/XbaI scar for RBS-CDS junctions</dcterms:description>
<dcterms:modified>2015-08-31T04:07:27Z</dcterms:modified>
<dcterms:title>scar</dcterms:title>
<sbol:displayId>BBa_G0000</sbol:displayId>
<sbol:persistentIdentity rdf:resource="https://synbiohub.org/public/igem/BBa_G0000"/>
<sbol:role rdf:resource="http://identifiers.org/so/SO:0000110"/>
<sbol:role rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#partType/DNA"/>
<sbol:sequence
        rdf:resource="https://synbiohub.org/public/igem/BBa_G0000_sequence/1"/>
<sbol:type rdf:resource="http://www.biopax.org/release/biopax-level3.owl#DnaRegion"/>
<sbol:version>1</sbol:version>
<igem:discontinued>false</igem:discontinued>
<igem:dominant>true</igem:dominant>
<igem:experience
        rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#experience/None"/>
<igem:group_u_list>_41_</igem:group_u_list>
<igem:m_user_id>0</igem:m_user_id>
<igem:owner_id>126</igem:owner_id>
<igem:owning_group_id>162</igem:owning_group_id>
<igem:sampleStatus>Not in stock</igem:sampleStatus>
<igem:status
        rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#status/Unavailable"/>
<sbh:bookmark>false</sbh:bookmark>
<sbh:mutableDescription>
    This is the sequence of the SpeI/XbaI scar for RBS-CDS junctions in
    BioBricks standard assembly.
</sbh:mutableDescription>
<sbh:mutableNotes>
    This is a shorter scar to ensure proper spacing between the RBS and CDS.
</sbh:mutableNotes>
<sbh:mutableProvenance>SpeI/XbaI scar</sbh:mutableProvenance>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/james"/>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/myers"/>
<sbh:star>false</sbh:star>
<sbh:topLevel rdf:resource="https://synbiohub.org/public/igem/BBa_G0000/1"/>
<prov:wasDerivedFrom rdf:resource="http://parts.igem.org/Part:BBa_G0000"/>
<prov:wasGeneratedBy rdf:resource="https://synbiohub.org/public/igem/igem2sbol/1"/>
</sbol:ComponentDefinition>
<sbol:Sequence rdf:about="https://synbiohub.org/public/igem/BBa_G0000_sequence/1">
<sbol:displayId>BBa_G0000_sequence</sbol:displayId>
<sbol:elements>tactag</sbol:elements>
<sbol:encoding rdf:resource="http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html"/>
<sbol:persistentIdentity
        rdf:resource="https://synbiohub.org/public/igem/BBa_G0000_sequence"/>
<sbol:version>1</sbol:version>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/james"/>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/myers"/>
<sbh:topLevel rdf:resource="https://synbiohub.org/public/igem/BBa_G0000_sequence/1"/>
<prov:wasDerivedFrom rdf:resource="http://parts.igem.org/Part:BBa_G0000"/>
<prov:wasGeneratedBy rdf:resource="https://synbiohub.org/public/igem/igem2sbol/1"/>
</sbol:Sequence>
<sbol:ComponentDefinition rdf:about="https://synbiohub.org/public/igem/BBa_G0002/1">
<dc:creator>Reshma Shetty</dc:creator>
<dcterms:created>2007-02-26T12:00:00Z</dcterms:created>
<dcterms:description>SpeI/XbaI mixed site</dcterms:description>
<dcterms:modified>2015-08-31T04:07:27Z</dcterms:modified>
<dcterms:title>SX scar</dcterms:title>
<sbol:displayId>BBa_G0002</sbol:displayId>
<sbol:persistentIdentity rdf:resource="https://synbiohub.org/public/igem/BBa_G0002"/>
<sbol:role rdf:resource="http://identifiers.org/so/SO:0000110"/>
<sbol:role rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#partType/DNA"/>
<sbol:sequence rdf:resource="https://synbiohub.org/public/igem/BBa_G0002_sequence/1"/>
<sbol:type rdf:resource="http://www.biopax.org/release/biopax-level3.owl#DnaRegion"/>
<sbol:version>1</sbol:version>
<igem:discontinued>false</igem:discontinued>
<igem:dominant>true</igem:dominant>
<igem:experience
        rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#experience/None"/>
<igem:group_u_list>_41_</igem:group_u_list>
<igem:m_user_id>0</igem:m_user_id>
<igem:owner_id>126</igem:owner_id>
<igem:owning_group_id>70</igem:owning_group_id>
<igem:sampleStatus>Not in stock</igem:sampleStatus>
<igem:status
        rdf:resource="http://wiki.synbiohub.org/wiki/Terms/igem#status/Unavailable"/>
<sbh:bookmark>false</sbh:bookmark>
<sbh:mutableDescription>
    XbaI/SpeI mixed site.  Simply used to aid in entry of parts into the registry.
</sbh:mutableDescription>
<sbh:mutableNotes>None.</sbh:mutableNotes>
<sbh:mutableProvenance>XbaI and SpeI sites</sbh:mutableProvenance>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/james"/>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/myers"/>
<sbh:star>false</sbh:star>
<sbh:topLevel rdf:resource="https://synbiohub.org/public/igem/BBa_G0002/1"/>
<prov:wasDerivedFrom rdf:resource="http://parts.igem.org/Part:BBa_G0002"/>
<prov:wasGeneratedBy rdf:resource="https://synbiohub.org/public/igem/igem2sbol/1"/>
</sbol:ComponentDefinition>
<sbol:Sequence rdf:about="https://synbiohub.org/public/igem/BBa_G0002_sequence/1">
<sbol:displayId>BBa_G0002_sequence</sbol:displayId>
<sbol:elements>tactagag</sbol:elements>
<sbol:encoding rdf:resource="http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html"/>
<sbol:persistentIdentity
        rdf:resource="https://synbiohub.org/public/igem/BBa_G0002_sequence"/>
<sbol:version>1</sbol:version>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/james"/>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/myers"/>
<sbh:topLevel rdf:resource="https://synbiohub.org/public/igem/BBa_G0002_sequence/1"/>
<prov:wasDerivedFrom rdf:resource="http://parts.igem.org/Part:BBa_G0002"/>
<prov:wasGeneratedBy rdf:resource="https://synbiohub.org/public/igem/igem2sbol/1"/>
</sbol:Sequence>
<prov:Activity rdf:about="https://synbiohub.org/public/igem/igem2sbol/1">
<dc:creator>Chris J. Myers</dc:creator>
<dc:creator>James Alastair McLaughlin</dc:creator>
<dcterms:description>Conversion of the iGEM parts registry to SBOL2.1
</dcterms:description>
<dcterms:title>iGEM to SBOL conversion</dcterms:title>
<sbol:displayId>igem2sbol</sbol:displayId>
<sbol:persistentIdentity rdf:resource="https://synbiohub.org/public/igem/igem2sbol"/>
<sbol:version>1</sbol:version>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/james"/>
<sbh:ownedBy rdf:resource="https://synbiohub.org/user/myers"/>
<sbh:topLevel rdf:resource="https://synbiohub.org/public/igem/igem2sbol/1"/>
<prov:endedAtTime>2017-03-06T15:00:00.000Z</prov:endedAtTime>
</prov:Activity>
</rdf:RDF>'''


def IGEM_STANDARD_ASSEMBLY(parts_list):
    if not all(type(part) is ComponentDefinition for part in parts_list):
        raise TypeError()
    doc = parts_list[0].doc
    G0000_uri = 'https://synbiohub.org/public/igem/BBa_G0000/1'
    G0000_seq_uri = 'https://synbiohub.org/public/igem/BBa_G0000_sequence/1'
    G0002_uri = 'https://synbiohub.org/public/igem/BBa_G0002/1'
    G0002_seq_uri = 'https://synbiohub.org/public/igem/BBa_G0002_sequence/1'
    if not (G0000_uri in doc.componentDefinitions and
            G0002_uri in doc.componentDefinitions and
            G0000_seq_uri in doc.sequences and G0002_seq_uri in doc.sequences):
        doc.appendString(igem_assembly_scars, overwrite=True)

    G0000 = doc.componentDefinitions[G0000_uri]
    G0002 = doc.componentDefinitions[G0002_uri]
    new_parts_list = []
    for upstream, downstream in zip(parts_list[:-1], parts_list[1:]):
        new_parts_list.append(upstream)
        if SO_RBS in upstream.roles and SO_CDS in downstream.roles:
            new_parts_list.append(G0000)
        else:
            new_parts_list.append(G0002)
    new_parts_list.append(downstream)
    return new_parts_list



import sbol2

# This example uses the SBOL Validator (http://synbiodex.github.io/SBOL-Validator/)
# to convert an SBOL document to a GenBank file.

# Create a new Document and read the SBOL file
doc = sbol2.Document('data/BBa_I0462.xml')

# Export the document in 'GenBank' format.
# Write the output to the destination file.
doc.exportToFormat('GenBank', 'BBa_I0462.gb')



import math

import sbol2


class MySampleSequence(sbol2.Identified):

    RDF_TYPE = 'http://example.org#MySampleSequence'

    def __init__(self, uri='example', letters='gcat'):
        super().__init__(uri=uri,
                         type_uri=MySample.RDF_TYPE)
        self.letters = sbol2.TextProperty(self, 'http://example.org#MyLetters', 1, 1,
                                          letters)


class MySample(sbol2.TopLevel):

    RDF_TYPE = 'http://example.org#MySample'

    def __init__(self, uri=None):
        super().__init__(uri=uri,
                         type_uri=MySample.RDF_TYPE)
        self.sequences = sbol2.OwnedObject(self, sbol2.SBOL_SEQUENCE_PROPERTY,
                                           MySampleSequence, 0, math.inf)


doc = sbol2.Document()
seq = sbol2.Sequence('seq1', 'cattag')
cd = sbol2.ComponentDefinition('cd1')
doc.add(cd)
cd.sequence = seq
print(doc.writeString())

sample_seq = MySampleSequence('myseq1', seq.elements)
sample = MySample('mysample1')
sample.sequences = [sample_seq]
doc.add(sample)
print(doc.writeString())



from .toplevel import *
from .constants import *
from .implementation import Implementation
from .experiment import ExperimentalData
from .collection import Collection


class Design(TopLevel):
    """This class represents a biological Design.
    A Design is a conceptual representation of a biological system
    that a synthetic biologist intends to build.
    A Design is the first object created in libSBOL's formalized
    Design-Build-Test-Analysis workflow."""

    def __init__(self, uri="example", version=VERSION_STRING,
                 structure=None, function=None):
        """Constructs a new Design.

        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled,
        then this argument is simply the displayId
        for the new object and a full URI will automatically be constructed.
        :param version: An arbitrary version string.
        If SBOLCompliance is enabled, this should be
        a Maven version string of the form "major.minor.patch".
        :param structure: A ComponentDefinition representing
        the structural aspects of the Design.
        :param function: A ModuleDefiniition representing
        the functional aspects of the Design.
        """
        super().__init__(SYSBIO_DESIGN, uri, version)
        self.characterization = ReferencedObject(self, SYSBIO_URI +
                                                 "#characterization",
                                                 SYSBIO_ANALYSIS, '0', '*',
                                                 [libsbol_rule_2])
        self.structure = OwnedObject(self, SBOL_COMPONENT_DEFINITION,
                                     '1', '1', [libsbol_rule_3], structure)
        self.function = OwnedObject(self, SBOL_MODULE_DEFINITION, '1', '1',
                                    [libsbol_rule_4], function)
        self._structure = ReferencedObject(self, SYSBIO_URI + "#_structure",
                                           SBOL_COMPONENT_DEFINITION,
                                           '1', '1', [], structure.identity)
        self._function = ReferencedObject(self, SYSBIO_URI + "#_function",
                                          SBOL_MODULE_DEFINITION, '1', '1',
                                          [], function.identity)

    def generateDesign(self, uri, agent=None, plan=None, usages=None):
        raise NotImplementedError("Not yet implemented")


class Build(Implementation):
    """A Build is a realization of a Design. For practical purposes,
    a Build can represent a biological clone,
    a plasmid, or other laboratory sample. For a given Design,
    there may be multiple Builds realized in the lab.
    A Build represents the second step
    in libSBOL's formalized Design-Build-Test-Analyze workflow."""

    def __init__(self, uri=URIRef('example'), version=VERSION_STRING, structure=None, behavior=None):
        super().__init__(uri, version)
        self.design = ReferencedObject(self, SYSBIO_URI + "#design", SYSBIO_DESIGN, '0', '1', [libsbol_rule_8])
        self.structure = OwnedObject(self, SBOL_COMPONENT_DEFINITION, '1', '1', [libsbol_rule_5], structure)
        self.behavior = OwnedObject(self, SBOL_MODULE_DEFINITION, '1', '1', [libsbol_rule_6], behavior)
        self._structure = ReferencedObject(self, SYSBIO_URI + "#_structure", SBOL_MODULE_DEFINITION, '1', '1', [], structure.identity)
        self._behavior = ReferencedObject(self, SBOL_URI + "#built", SBOL_MODULE_DEFINITION, '1', '1', [], behavior.identity)
        self._sysbio_type = URIProperty(self, SYSBIO_URI + "#type", '1', '1', [], SYSBIO_BUILD)

    @property
    def sysbio_type(self):
        return self._sysbio_type.value

    @sysbio_type.setter
    def sysbio_type(self, new_sysbio_type):
        self._sysbio_type.set(new_sysbio_type)


class Test(ExperimentalData):
    def __init__(self, uri=URIRef('example'), version=VERSION_STRING):
        super().__init__(uri, version)
        self.samples = ReferencedObject(self, SYSBIO_URI + "#samples", SBOL_IMPLEMENTATION, '0', '*', [libsbol_rule_9])
        self.dataFiles = ReferencedObject(self, SBOL_ATTACHMENTS, SBOL_ATTACHMENT, '0', '*', [])


class Analysis(TopLevel):
    def __init__(self, uri=URIRef('example'), version=VERSION_STRING):
        super().__init__(SYSBIO_ANALYSIS, uri, version)
        self.rawData = ReferencedObject(self, SYSBIO_URI + "#rawData", SBOL_EXPERIMENTAL_DATA, '0', '1', [libsbol_rule_10])
        self.dataFiles = ReferencedObject(self, SBOL_ATTACHMENTS, SBOL_ATTACHMENT, '0', '*', [])
        self.consensusSequence = OwnedObject(self, SYSBIO_URI + "#consensusSequence", '0', '1', [])
        self.fittedModel = OwnedObject(self, SYSBIO_URI + "#model", '0', '1', [])
        self.dataSheet = ReferencedObject(self, SYSBIO_URI + "#dataSheet", SBOL_ATTACHMENT, '0', '1', [])
        self._consensusSequence = OwnedObject(self, SYSBIO_URI + "#consensusSequence", SBOL_SEQUENCE, '0', '1', [])
        self._fittedModel = OwnedObject(self, SYSBIO_URI + "#model", SBOL_MODEL, '0', '1', [])

    def verifyTarget(self, consensus_sequence):
        raise NotImplementedError('Not yet implemented')

    def reportIdentity(self):
        raise NotImplementedError('Not yet implemented')

    def reportCoverage(self):
        raise NotImplementedError('Not yet implemented')

    def reportAmbiguity(self):
        raise NotImplementedError('Not yet implemented')

    def generate_analysis(self, uri, agent=None, plan=None, usages=None):
        raise NotImplementedError('Not yet implemented')


class SampleRoster(Collection):
    """A SampleRoster is a container used to group Builds that are included in an experiment together.

    A SampleRoster can be used to generate a Test in a Design-Build-Test-Learn workflow."""

    def __init__(self, uri=URIRef('example'), version=VERSION_STRING):
        super().__init__(uri, version)
        self.samples = ReferencedObject(self, SBOL_MEMBERS,
                                        SBOL_IMPLEMENTATION, '0', '*',
                                        [libsbol_rule_15])
        self._sysbio_type = URIProperty(self, SYSBIO_URI + "#type", '1', '1',
                                        [], SYSBIO_URI + "#SampleRoster")
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            self.identity = URIRef(Config.getHomespace() + '/SampleRoster/' +
                                   self.displayId + '/' + version)
            self.persistentIdentity = URIRef(Config.getHomespace() +
                                             '/SampleRoster/' + self.displayId)

    @property
    def sysbio_type(self):
        return self._sysbio_type.value

    @sysbio_type.setter
    def sysbio_type(self, new_sysbio_type):
        self._sysbio_type.set(new_sysbio_type)


class AliasedProperty:
    # TODO
    pass


class TranscriptionRepressionInteraction:
    # TODO
    pass


class SmallMoleculeInhibitionInteraction:
    # TODO
    pass


class GeneProductionInteraction:
    # TODO
    pass


class TranscriptionalActivationInteraction:
    # TODO
    pass


class SmallMoleculeActivationInteraction:
    # TODO
    pass


class EnzymeCatalysisInteraction:
    # TODO
    pass
